# File: src/samudra_ai/__init__.py
from .core import SamudraAI, SamudraAI2
from .data_loader import load_and_mask_dataset
from .trainer import prepare_training_data, plot_training_history
from .evaluator import evaluate_model, plot_spatial_comparison, mask_land, mask_ocean
from .preprocess_dcpp import preprocess_dcpp

__version__ = "1.3.3"
__all__ = [
    'SamudraAI',
    'SamudraAI2',
    'load_and_mask_dataset',
    'prepare_training_data',
    'plot_training_history',
    'evaluate_model',
    'plot_spatial_comparison',
    "preprocess_dcpp",
    'mask_land',
    'mask_ocean',
]


# File: src/samudra_ai/trainer.py
from sklearn.preprocessing import MinMaxScaler
from sklearn.model_selection import train_test_split
import numpy as np
import tensorflow as tf
import random
import matplotlib.pyplot as plt
import os

def prepare_training_data(x_data, y_data, time_seq, seed):
    np.random.seed(seed)
    random.seed(seed)
    tf.random.set_seed(seed)

    min_samples = min(x_data.shape[0], y_data.shape[0])
    x_np = np.nan_to_num(x_data.isel(time=slice(0, min_samples)).values)
    y_np = np.nan_to_num(y_data.isel(time=slice(0, min_samples)).values)

    scaler_X = MinMaxScaler()
    X_scaled = scaler_X.fit_transform(x_np.reshape(x_np.shape[0], -1)).reshape(x_np.shape)
    scaler_y = MinMaxScaler(feature_range=(-1, 1))
    y_scaled = scaler_y.fit_transform(y_np.reshape(y_np.shape[0], -1)).reshape(y_np.shape)

    X_seq = np.array([X_scaled[i:i+time_seq] for i in range(len(X_scaled) - time_seq)])
    y_seq = np.array([y_scaled[i+time_seq-1] for i in range(len(y_scaled) - time_seq)])

    X_train, X_test, y_train, y_test = train_test_split(
        X_seq[..., np.newaxis], y_seq, test_size=0.2, random_state=seed, shuffle=False)

    return X_train, X_test, y_train, y_test, scaler_X, scaler_y

def plot_training_history(history, output_dir=None):
    if output_dir: os.makedirs(output_dir, exist_ok=True)
    plt.figure(figsize=(10, 6))
    plt.plot(history.history['loss'], label='Train Loss')
    plt.plot(history.history['val_loss'], label='Val Loss')
    plt.title("Kurva Loss Training")
    plt.xlabel("Epoch")
    plt.ylabel("Loss")
    plt.legend()
    if output_dir:
        plt.savefig(os.path.join(output_dir, "loss_curve.png"))
    plt.show()

    plt.figure(figsize=(10, 6))
    plt.plot(history.history['mae'], label='Train MAE')
    plt.plot(history.history['val_mae'], label='Val MAE')
    plt.title("Kurva MAE Training")
    plt.xlabel("Epoch")
    plt.ylabel("MAE")
    plt.legend()
    if output_dir:
        plt.savefig(os.path.join(output_dir, "mae_curve.png"))
    plt.show()


# File: src/samudra_ai/preprocess_dcpp.py
import xarray as xr
import pandas as pd
import numpy as np

def preprocess_dcpp(file_path: str, var_name: str, lat_range=None, lon_range=None, time_range=None):
    print(f"-> Preprocessing DCPP file: {file_path} (var: {var_name})")

    ds = xr.open_dataset(file_path, engine="h5netcdf", decode_times=True)

    if var_name not in ds:
        raise ValueError(f"Variabel {var_name} tidak ada dalam file {file_path}")

    da = ds[var_name]

    # === Hapus dimensi tidak relevan ===
    for dim in ["initial_time", "ensemble", "member", "zlev"]:
        if dim in da.dims:
            da = da.isel({dim: 0}).squeeze(dim)

    # Kalau masih ada dimensi extra selain time/lat/lon → ambil indeks pertama
    while len(da.dims) > 3:
        for d in da.dims:
            if d not in ["time", "lat", "lon"]:
                da = da.isel({d: 0}).squeeze(d)

    # === Deteksi & rename dimensi waktu ===
    time_candidates = ["time", "valid_time", "date", "Time", "t"]
    detected_time = next((t for t in time_candidates if t in da.dims or t in da.coords), None)
    if not detected_time:
        raise ValueError("Dimensi waktu tidak ditemukan di file DCPP.")

    if detected_time != "time":
        da = da.rename({detected_time: "time"})

    # === Slicing waktu kalau diberikan ===
    if time_range:
        try:
            start_dt = pd.to_datetime(time_range[0])
            end_dt = pd.to_datetime(time_range[1])
        except Exception:
            raise ValueError("Format time_range harus bisa dibaca pandas, misal '1981-01-01'")
        if start_dt >= end_dt:
            raise ValueError("time_range: tanggal awal harus lebih kecil dari tanggal akhir")

        time_type = type(da.time.values[0])
        if "cftime" in str(time_type):
            start_time = time_type(start_dt.year, start_dt.month, start_dt.day)
            end_time = time_type(end_dt.year, end_dt.month, end_dt.day)
        else:
            start_time = np.datetime64(start_dt, "D")
            end_time = np.datetime64(end_dt, "D")

        start_time_sel = da.time.sel(time=start_time, method="nearest").values
        end_time_sel = da.time.sel(time=end_time, method="nearest").values
        da = da.sel(time=slice(start_time_sel, end_time_sel))

    # === Slicing lat/lon kalau diberikan ===
    if lat_range and lon_range:
        lat_names = ["lat", "latitude", "j", "y"]
        lon_names = ["lon", "longitude", "i", "x"]
        detected_lat = next((lat for lat in lat_names if lat in da.dims), None)
        detected_lon = next((lon for lon in lon_names if lon in da.dims), None)
        if detected_lat and detected_lon:
            da = da.sel({detected_lat: slice(*lat_range), detected_lon: slice(*lon_range)})

    # === Pastikan urutan dimensi standar ===
    if set(["time", "lat", "lon"]).issubset(set(da.dims)):
        da = da.transpose("time", "lat", "lon")

    return da



# File: src/samudra_ai/core.py
import os
import json
import joblib
import logging
import xarray as xr
import numpy as np
from tensorflow.keras.models import load_model as keras_load_model
from tensorflow.keras.layers import LeakyReLU
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.callbacks import EarlyStopping, ReduceLROnPlateau

from .models import build_cnn_bilstm, ModelConfig
from .model2 import build_convlstm2d, ModelConfig2
from .utils import NumpyEncoder, save_to_netcdf
from .data_loader import load_and_mask_dataset
from .trainer import prepare_training_data, plot_training_history
from .evaluator import evaluate_model, plot_spatial_comparison

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

class SamudraAI:
    def __init__(self, time_seq: int = 9, lstm_units: int = 64, learning_rate: float = 1e-4, config: ModelConfig = None):
        self.time_seq = time_seq
        self.lstm_units = lstm_units
        self.learning_rate = learning_rate
        self.config = config
        self.model = None
        self.scaler_X = None
        self.scaler_y = None
        self.history = None
        self.obs_for_metadata = None

    def fit(self, x_data_hist: xr.DataArray, y_data_obs: xr.DataArray, epochs: int = 100, batch_size: int = 8, test_size: float = 0.2, seed: int = 42):
        logger.info("🧠 Memulai training model CNN-BiLSTM...")
        self.obs_for_metadata = y_data_obs.copy(deep=True)

        X_train, X_test, y_train, y_test, scaler_X, scaler_y = prepare_training_data(
            x_data_hist, y_data_obs, self.time_seq, seed
        )
        self.scaler_X = scaler_X
        self.scaler_y = scaler_y

        input_shape = X_train.shape[1:]
        output_shape = y_train.shape[1:3]

        # ✅ pakai config custom kalau ada, kalau tidak build default
        if self.config is None:
            self.config = ModelConfig(rnn_units=self.lstm_units, learning_rate=self.learning_rate)

        self.model = build_cnn_bilstm(input_shape, output_shape, self.config)
        self.model.summary()

        callbacks = [EarlyStopping(patience=10, restore_best_weights=True), ReduceLROnPlateau(patience=5, verbose=1)]
        self.history = self.model.fit(
            X_train, y_train, epochs=epochs, batch_size=batch_size,
            validation_data=(X_test, y_test), callbacks=callbacks, verbose=1
        )

        loss, mae = self.model.evaluate(X_test, y_test, verbose=0)
        logger.info(f"✅ Training selesai -> Final Val Loss (MSE): {loss:.4f}, Final Val MAE: {mae:.4f}")

    def correction(self, data_to_correct: xr.DataArray, save_path: str = None) -> xr.DataArray:
        if not all([self.model, self.scaler_X, self.scaler_y, self.obs_for_metadata is not None]):
            raise RuntimeError("Model belum dilatih. Jalankan .fit() terlebih dahulu.")

        logger.info(f"🔬 Mengoreksi data dengan shape: {data_to_correct.shape}...")

        X_future = data_to_correct.fillna(0).values
        X_future_scaled = self.scaler_X.transform(X_future.reshape(X_future.shape[0], -1)).reshape(X_future.shape)
        X_future_seq = np.array([X_future_scaled[i:i+self.time_seq] for i in range(len(X_future_scaled) - self.time_seq)])

        X_future_seq = X_future_seq[..., np.newaxis]  # ✅ bentuk ke 5D
        print("🧪 Input to model.predict:", X_future_seq.shape)
        print("🧪 Any NaN in input?:", np.isnan(X_future_seq).any())

        pred_scaled = self.model.predict(X_future_seq, batch_size=1).squeeze()

        flat_pred = pred_scaled.reshape(pred_scaled.shape[0], -1)

        # Ambil shape dari data observasi (tanpa waktu)
        obs_shape = self.obs_for_metadata.isel(time=0).shape
        expected_size = np.prod(obs_shape)

        if flat_pred.shape[1] != expected_size:
            raise ValueError(
                f"Mismatch: Hasil prediksi ({flat_pred.shape[1]}) tidak cocok dengan target obs_ref "
                f"({expected_size}) berdasarkan shape {obs_shape}"
            )

        pred_original = self.scaler_y.inverse_transform(flat_pred).reshape((pred_scaled.shape[0], *obs_shape))

        obs_mask = self.obs_for_metadata.isel(time=0).notnull()
        lat_dim_name = next((d for d in obs_mask.dims if 'lat' in d.lower()), 'lat')
        lon_dim_name = next((d for d in obs_mask.dims if 'lon' in d.lower()), 'lon')

        valid_time = data_to_correct.time.values[self.time_seq - 1 : self.time_seq - 1 + len(pred_original)]

        predicted_da = xr.DataArray(
            pred_original,
            coords={
                "time": valid_time,
                lat_dim_name: obs_mask.coords[lat_dim_name],
                lon_dim_name: obs_mask.coords[lon_dim_name]
            },
            dims=["time", lat_dim_name, lon_dim_name],
            name="corrected"
        )

        if predicted_da.shape[1:] == obs_mask.shape:
            predicted_masked = predicted_da.where(obs_mask, drop=False)
        else:
            logger.warning("Ukuran mask observasi tidak cocok. Koreksi akan disimpan tanpa masking akhir.")
            predicted_masked = predicted_da

        # Jika disediakan save_path, simpan otomatis
        if save_path:
            save_to_netcdf(predicted_masked, save_path)

        logger.info("✅ Koreksi selesai.")
        return predicted_masked

    def evaluate_and_plot(
        self,
        raw_gcm_data: xr.DataArray,
        ref_data: xr.DataArray,
        var_name_ref: str,
        output_dir: str = "hasil_evaluasi",
        save_corrected_path: str = None,
        mask_type: str = None):
        
        corrected_data = self.correction(raw_gcm_data)

        # Debug log hasil prediksi
        try:
            min_val = np.nanmin(corrected_data.values)
            max_val = np.nanmax(corrected_data.values)
            print(f"✅ Hasil koreksi -> min: {min_val:.4f}, max: {max_val:.4f}, shape: {corrected_data.shape}")
        except Exception as e:
            print("⚠️ Gagal menghitung min/max hasil koreksi:", e)

        # Auto-trim obs agar cocok dengan hasil koreksi
        if corrected_data.sizes['time'] != ref_data.sizes['time']:
            start_index = self.time_seq - 1
            ref_data = ref_data.isel(time=slice(start_index, start_index + corrected_data.sizes['time']))

        # Simpan hasil koreksi jika diminta
        if save_corrected_path:
            save_to_netcdf(corrected_data, save_corrected_path)

        # === Plot spasial perbandingan ===
        spatial_path = plot_spatial_comparison(
            corrected_da=corrected_data,
            raw_gcm_da=raw_gcm_data,
            ref_da=ref_data,
            var_name=var_name_ref,
            output_dir=output_dir,
            units=ref_data.attrs.get("units", ""),
            mask_type=mask_type 
        )
        logger.info(f"🌍 Peta spasial disimpan ke: {spatial_path}")

        return evaluate_model(ref_data, raw_gcm_data, corrected_data, var_name_ref, output_dir)

    def plot_history(self, output_dir: str = None):
        if not self.history:
            raise ValueError("Model belum dilatih.")
        plot_training_history(self.history, output_dir)

    def save(self, path: str):
        logger.info(f"💾 Menyimpan model ke direktori: {path}...")
        os.makedirs(path, exist_ok=True)

        # Simpan model dan scaler
        self.model.save(os.path.join(path, "model.keras"))
        joblib.dump(self.scaler_X, os.path.join(path, "scaler_X.gz"))
        joblib.dump(self.scaler_y, os.path.join(path, "scaler_y.gz"))

        # Simpan metadata observasi dengan backend yang lebih stabil
        try:
            self.obs_for_metadata.to_netcdf(
                os.path.join(path, "obs_metadata.nc"),
                engine="h5netcdf"
            )
        except Exception as e:
            logger.warning(f"❗ Gagal menyimpan obs_metadata.nc: {e}")

        # Simpan konfigurasi model
        config = {
            'time_seq': self.time_seq,
            'lstm_units': self.lstm_units,
            'learning_rate': self.learning_rate
        }
        with open(os.path.join(path, "config.json"), 'w') as f:
            json.dump(config, f, cls=NumpyEncoder)

        logger.info(f"✅ Model dan komponen berhasil disimpan.")

    @classmethod
    def load(cls, path: str):
        logger.info(f"🔄 Memuat model dari direktori: {path}...")
        required_files = ["config.json", "model.keras", "scaler_X.gz", "scaler_y.gz", "obs_metadata.nc"]
        for fname in required_files:
            full_path = os.path.join(path, fname)
            if not os.path.exists(full_path):
                raise FileNotFoundError(f"File tidak ditemukan: {full_path}")

        with open(os.path.join(path, "config.json"), 'r') as f:
            config = json.load(f)
        instance = cls(**config)
        instance.model = keras_load_model(
            os.path.join(path, "model.keras"),
            custom_objects={"LeakyReLU": LeakyReLU},
            compile=False)
        instance.scaler_X = joblib.load(os.path.join(path, "scaler_X.gz"))
        instance.scaler_y = joblib.load(os.path.join(path, "scaler_y.gz"))
        instance.obs_for_metadata = xr.open_dataarray(os.path.join(path, "obs_metadata.nc"))
        instance.model.compile(
            optimizer=Adam(learning_rate=instance.learning_rate),
            loss='mse',
            metrics=['mae'])
        logger.info(f"✅ Model dan komponen berhasil dimuat.")
        return instance
    
class SamudraAI2:
    def __init__(self, time_seq: int = 9, config: ModelConfig2 = None):
        self.time_seq = time_seq
        self.config = config
        self.model = None
        self.scaler_X = None
        self.scaler_y = None
        self.history = None
        self.obs_for_metadata = None

    def fit(self, x_data_hist, y_data_obs, epochs: int = None, batch_size: int = None, test_size: float = 0.2, seed: int = 42):
        logger.info("🧠 Memulai training model ConvLSTM2D...")
        self.obs_for_metadata = y_data_obs.copy(deep=True)

        X_train, X_test, y_train, y_test, scaler_X, scaler_y = prepare_training_data(
            x_data_hist, y_data_obs, self.time_seq, seed
        )
        self.scaler_X, self.scaler_y = scaler_X, scaler_y

        input_shape = X_train.shape[1:]    # (time_seq, lat, lon, 1)
        output_shape = y_train.shape[1:3]  # (lat, lon)

        if self.config is None:
            self.config = ModelConfig2()

        self.model = build_convlstm2d(input_shape, output_shape, self.config)
        self.model.summary()

        callbacks = [EarlyStopping(patience=10, restore_best_weights=True), ReduceLROnPlateau(patience=5, verbose=1)]
        self.history = self.model.fit(
            X_train, y_train,
            epochs=epochs or self.config.epochs,
            batch_size=batch_size or self.config.batch_size,
            validation_data=(X_test, y_test),
            callbacks=callbacks, verbose=1
        )

    def correction(self, data_to_correct: xr.DataArray, save_path: str = None) -> xr.DataArray:
        if not all([self.model, self.scaler_X, self.scaler_y, self.obs_for_metadata is not None]):
            raise RuntimeError("Model belum dilatih. Jalankan .fit() terlebih dahulu.")

        logger.info(f"🔬 Mengoreksi data dengan shape: {data_to_correct.shape}...")

        X_future = data_to_correct.fillna(0).values
        X_future_scaled = self.scaler_X.transform(X_future.reshape(X_future.shape[0], -1)).reshape(X_future.shape)
        X_future_seq = np.array([X_future_scaled[i:i+self.time_seq] for i in range(len(X_future_scaled) - self.time_seq)])

        X_future_seq = X_future_seq[..., np.newaxis]  # ✅ bentuk ke 5D
        print("🧪 Input to model.predict:", X_future_seq.shape)
        print("🧪 Any NaN in input?:", np.isnan(X_future_seq).any())

        pred_scaled = self.model.predict(X_future_seq, batch_size=1).squeeze()

        flat_pred = pred_scaled.reshape(pred_scaled.shape[0], -1)

        # Ambil shape dari data observasi (tanpa waktu)
        obs_shape = self.obs_for_metadata.isel(time=0).shape
        expected_size = np.prod(obs_shape)

        if flat_pred.shape[1] != expected_size:
            raise ValueError(
                f"Mismatch: Hasil prediksi ({flat_pred.shape[1]}) tidak cocok dengan target obs_ref "
                f"({expected_size}) berdasarkan shape {obs_shape}"
            )

        pred_original = self.scaler_y.inverse_transform(flat_pred).reshape((pred_scaled.shape[0], *obs_shape))

        obs_mask = self.obs_for_metadata.isel(time=0).notnull()
        lat_dim_name = next((d for d in obs_mask.dims if 'lat' in d.lower()), 'lat')
        lon_dim_name = next((d for d in obs_mask.dims if 'lon' in d.lower()), 'lon')

        valid_time = data_to_correct.time.values[self.time_seq - 1 : self.time_seq - 1 + len(pred_original)]

        predicted_da = xr.DataArray(
            pred_original,
            coords={
                "time": valid_time,
                lat_dim_name: obs_mask.coords[lat_dim_name],
                lon_dim_name: obs_mask.coords[lon_dim_name]
            },
            dims=["time", lat_dim_name, lon_dim_name],
            name="corrected"
        )

        if predicted_da.shape[1:] == obs_mask.shape:
            predicted_masked = predicted_da.where(obs_mask, drop=False)
        else:
            logger.warning("Ukuran mask observasi tidak cocok. Koreksi akan disimpan tanpa masking akhir.")
            predicted_masked = predicted_da

        # Jika disediakan save_path, simpan otomatis
        if save_path:
            save_to_netcdf(predicted_masked, save_path)

        logger.info("✅ Koreksi selesai.")
        return predicted_masked

    def evaluate_and_plot(
        self,
        raw_gcm_data: xr.DataArray,
        ref_data: xr.DataArray,
        var_name_ref: str,
        output_dir: str = "hasil_evaluasi",
        save_corrected_path: str = None,
        mask_type: str = None):

        logger.info("📊 Evaluasi model dengan data referensi...")
        
        corrected_data = self.correction(raw_gcm_data)

        # Debug log hasil prediksi
        try:
            min_val = np.nanmin(corrected_data.values)
            max_val = np.nanmax(corrected_data.values)
            print(f"✅ Hasil koreksi -> min: {min_val:.4f}, max: {max_val:.4f}, shape: {corrected_data.shape}")
        except Exception as e:
            print("⚠️ Gagal menghitung min/max hasil koreksi:", e)

        # Auto-trim obs agar cocok dengan hasil koreksi
        if corrected_data.sizes['time'] != ref_data.sizes['time']:
            start_index = self.time_seq - 1
            ref_data = ref_data.isel(time=slice(start_index, start_index + corrected_data.sizes['time']))

        # Simpan hasil koreksi jika diminta
        if save_corrected_path:
            save_to_netcdf(corrected_data, save_corrected_path)

        # === Plot spasial perbandingan ===
        spatial_path = plot_spatial_comparison(
            corrected_da=corrected_data,
            raw_gcm_da=raw_gcm_data,
            ref_da=ref_data,
            var_name=var_name_ref,
            output_dir=output_dir,
            units=ref_data.attrs.get("units", ""),
            mask_type=mask_type 
        )
        logger.info(f"🌍 Peta spasial disimpan ke: {spatial_path}")

        return evaluate_model(ref_data, raw_gcm_data, corrected_data, var_name_ref, output_dir)

    def plot_history(self, output_dir: str = None):
        if not self.history:
            raise ValueError("Model belum dilatih.")
        plot_training_history(self.history, output_dir)

    def save(self, path: str):
        logger.info(f"💾 Menyimpan model ke direktori: {path}...")
        os.makedirs(path, exist_ok=True)

        # Simpan model dan scaler
        self.model.save(os.path.join(path, "model2.keras"))
        joblib.dump(self.scaler_X, os.path.join(path, "scaler_X2.gz"))
        joblib.dump(self.scaler_y, os.path.join(path, "scaler_y2.gz"))

        # Simpan metadata observasi dengan backend yang lebih stabil
        try:
            self.obs_for_metadata.to_netcdf(
                os.path.join(path, "obs_metadata2.nc"),
                engine="h5netcdf"
            )
        except Exception as e:
            logger.warning(f"❗ Gagal menyimpan obs_metadata.nc: {e}")

        # Simpan konfigurasi model
        config = {
            'time_seq': self.time_seq,
            'lstm_units': self.lstm_units,
            'learning_rate': self.learning_rate
        }
        with open(os.path.join(path, "config2.json"), 'w') as f:
            json.dump(config, f, cls=NumpyEncoder)

        logger.info(f"✅ Model dan komponen berhasil disimpan.")

    @classmethod
    def load(cls, path: str):
        logger.info(f"🔄 Memuat model dari direktori: {path}...")
        required_files = ["config2.json", "model2.keras", "scaler_X2.gz", "scaler_y2.gz", "obs_metadata2.nc"]
        for fname in required_files:
            full_path = os.path.join(path, fname)
            if not os.path.exists(full_path):
                raise FileNotFoundError(f"File tidak ditemukan: {full_path}")

        with open(os.path.join(path, "config.json"), 'r') as f:
            config = json.load(f)
        instance = cls(**config)
        instance.model = keras_load_model(
            os.path.join(path, "model.keras"),
            custom_objects={"LeakyReLU": LeakyReLU},
            compile=False)
        instance.scaler_X = joblib.load(os.path.join(path, "scaler_X.gz"))
        instance.scaler_y = joblib.load(os.path.join(path, "scaler_y.gz"))
        instance.obs_for_metadata = xr.open_dataarray(os.path.join(path, "obs_metadata.nc"))
        instance.model.compile(
            optimizer=Adam(learning_rate=instance.learning_rate),
            loss='mse',
            metrics=['mae'])
        logger.info(f"✅ Model CONVLSTM2D dan komponen berhasil dimuat.")
        return instance


# File: src/samudra_ai/evaluator.py
import pandas as pd
import numpy as np
import os
import matplotlib.pyplot as plt
import seaborn as sns
import xarray as xr
import cartopy.crs as ccrs
import cartopy.feature as cfeature
from shapely.geometry import Point
from shapely.prepared import prep
from scipy.interpolate import interpn
from .utils import compute_metrics, standardize_dims

def evaluate_model(ref_data, raw_gcm_data, corrected_data, var_name, output_dir):
    print("\n🔎 Memulai proses evaluasi dan plotting otomatis...")

    # Standarisasi dimensi time-lat-lon
    ref_data = standardize_dims(ref_data)
    raw_gcm_data = standardize_dims(raw_gcm_data)
    corrected_data = standardize_dims(corrected_data)

    # Potong ke time length minimum
    time_len = min(len(ref_data.time), len(raw_gcm_data.time), len(corrected_data.time))
    ref_sliced = ref_data.isel(time=slice(0, time_len))
    raw_sliced = raw_gcm_data.isel(time=slice(0, time_len))
    corr_sliced = corrected_data.isel(time=slice(0, time_len))

    # Ambil grid target dari referensi
    target_lat = ref_sliced['lat'].values
    target_lon = ref_sliced['lon'].values
    target_grid_lat, target_grid_lon = np.meshgrid(target_lat, target_lon, indexing='ij')
    target_points = np.array([target_grid_lat.ravel(), target_grid_lon.ravel()]).T

    def manual_interp(source_da, target_da):
        # Pastikan dimensi [time, lat, lon]
        if source_da.dims[-2:] != ("lat", "lon"):
            source_da = source_da.transpose("time", "lat", "lon")

        source_lat = source_da['lat'].values
        source_lon = source_da['lon'].values

        # Validasi monoton
        if not (np.all(np.diff(source_lat) > 0) or np.all(np.diff(source_lat) < 0)):
            raise ValueError("Latitude harus monoton. Urutkan ulang dimensi 'lat'.")
        if not (np.all(np.diff(source_lon) > 0) or np.all(np.diff(source_lon) < 0)):
            raise ValueError("Longitude harus monoton. Urutkan ulang dimensi 'lon'.")

        source_points = (source_lat, source_lon)
        interpolated_values = np.array([
            interpn(
                source_points, source_da.isel(time=t).values, target_points,
                method='linear', bounds_error=False, fill_value=np.nan
            ).reshape(target_grid_lat.shape)
            for t in range(source_da.shape[0])
        ])
        # return xr.DataArray(interpolated_values, coords=ref_sliced.coords, dims=ref_sliced.dims)
        return xr.DataArray(
            interpolated_values,
            coords={"time": target_da["time"], "lat": target_da["lat"], "lon": target_da["lon"]},
            dims=["time", "lat", "lon"]
        )

    # Interpolasi raw GCM ke grid ref
    raw_aligned = manual_interp(raw_sliced, ref_sliced)

    # Rata-rata spasial jadi timeseries
    ref_series = ref_sliced.mean(dim=['lat', 'lon'])
    raw_series = raw_aligned.mean(dim=['lat', 'lon'])
    corr_series = corr_sliced.mean(dim=['lat', 'lon'])

    # Hitung metrik evaluasi
    metrics_raw = compute_metrics(ref_series.values, raw_series.values)
    metrics_corr = compute_metrics(ref_series.values, corr_series.values)
    results_df = pd.DataFrame([
        {'Source': 'GCM Asli', **dict(zip(['Correlation', 'RMSE', 'Bias', 'MAE'], metrics_raw))},
        {'Source': 'GCM Terkoreksi', **dict(zip(['Correlation', 'RMSE', 'Bias', 'MAE'], metrics_corr))},
    ])

    # Simpan summary metrics
    if output_dir:
        os.makedirs(output_dir, exist_ok=True)
        results_df.to_excel(os.path.join(output_dir, "summary_metrics.xlsx"), index=False)

    # Simpan timeseries per timestep
    df_line = pd.DataFrame({
        "Time": ref_series.time.values,
        "GCM Asli": raw_series.values,
        "GCM Terkoreksi": corr_series.values,
        "Reanalysis (Obs)": ref_series.values
    })

    if output_dir:
        df_line.to_excel(os.path.join(output_dir, "timeseries_comparison.xlsx"), index=False)

    # Plot timeseries
    plt.figure(figsize=(15, 6))
    sns.lineplot(data=df_line)
    plt.title(f"Grafik Perbandingan Time Series {var_name.upper()}")
    plt.grid(True)
    if output_dir:
        plt.savefig(os.path.join(output_dir, "timeseries_comparison.png"), dpi=300)
    plt.show()

    # Plot bar metrik
    df_melt = results_df.melt(id_vars=['Source'], var_name='Metric', value_name='Value')
    plt.figure(figsize=(10, 7))
    sns.barplot(data=df_melt, x='Metric', y='Value', hue='Source', palette='viridis')
    plt.title("Perbandingan Metrik Evaluasi")
    if output_dir:
        plt.savefig(os.path.join(output_dir, "metrics_barplot.png"), dpi=300)
    plt.show()

    return results_df, corr_sliced

def _create_land_mask(lat, lon, resolution="10m"):
    """Buat masker daratan berdasarkan NaturalEarthFeature."""
    land_feature = cfeature.NaturalEarthFeature("physical", "land", resolution)
    land_geom = list(land_feature.geometries())
    land_prep = [prep(g) for g in land_geom]

    mask = np.zeros((len(lat), len(lon)), dtype=bool)
    for i, la in enumerate(lat):
        for j, lo in enumerate(lon):
            pt = Point(lo, la)
            if any(g.contains(pt) for g in land_prep):
                mask[i, j] = True
    return mask

def mask_land(da, resolution="10m"):
    """Masking hanya daratan."""
    lat = da["lat"].values
    lon = da["lon"].values
    mask = _create_land_mask(lat, lon, resolution)
    return da.where(mask)


def mask_ocean(da, resolution="10m"):
    """Masking hanya lautan (kebalikan dari mask_land)."""
    lat = da["lat"].values
    lon = da["lon"].values
    mask = _create_land_mask(lat, lon, resolution)
    return da.where(~mask)

def plot_spatial_comparison(corrected_da, raw_gcm_da, ref_da, var_name, output_dir=None,
                            units="", mask_type=None):
    
    print("🌍 Membuat plot spasial perbandingan rata-rata...")

    # Hitung rata-rata sepanjang waktu
    out_mean = corrected_da.mean(dim="time")
    gcm_mean = raw_gcm_da.mean(dim="time")
    ref_mean = ref_da.mean(dim="time")

    # Masking opsional
    if mask_type == "land":
        out_mean, gcm_mean, ref_mean = mask_land(out_mean), mask_land(gcm_mean), mask_land(ref_mean)
    elif mask_type == "ocean":
        out_mean, gcm_mean, ref_mean = mask_ocean(out_mean), mask_ocean(gcm_mean), mask_ocean(ref_mean)

    datasets = [out_mean, gcm_mean, ref_mean]
    titles = ["Model Terkoreksi", "GCM Asli", "Observasi (Reanalysis)"]

    fig, axes = plt.subplots(1, 3, figsize=(18, 6), subplot_kw={"projection": ccrs.PlateCarree()})

    for ax, da, title in zip(axes, datasets, titles):
        im = ax.pcolormesh(
            da["lon"], da["lat"], da,
            cmap="coolwarm", transform=ccrs.PlateCarree()
        )
        ax.set_aspect('auto')
        ax.coastlines()
        ax.set_title(title)

    # === buat axis khusus untuk colorbar di bawah ===
    cbar_ax = fig.add_axes([0.25, 0.1, 0.5, 0.03])  
    # [left, bottom, width, height] dalam koordinat figure (0–1)
    cbar = fig.colorbar(im, cax=cbar_ax, orientation="horizontal")
    cbar.set_label(units)

    plt.suptitle(f"Peta Perbandingan Rata-rata Spasial {var_name.upper()}", fontsize=14)
    plt.tight_layout(rect=[0, 0.15, 1, 0.95])
    # fig.subplots_adjust(left=0.05, right=0.95, top=0.9, bottom=0.2, wspace=0.25)

    save_path = None
    if output_dir:
        os.makedirs(output_dir, exist_ok=True)
        save_path = os.path.join(output_dir, f"spatial_comparison_{var_name}_{mask_type or 'global'}.jpg")
        plt.savefig(save_path, dpi=300, bbox_inches="tight", format="jpg")
        print(f"✅ Peta spasial disimpan ke: {save_path}")

    plt.show()
    return save_path


import numpy as np
import xarray as xr
from samudra_ai import SamudraAI

# Fungsi pembuat data dummy
def create_dummy_data(time=20, lat=3, lon=4):
    coords = {
        "time": np.arange(time),
        "lat": np.linspace(-10, 10, lat),
        "lon": np.linspace(100, 120, lon),
    }
    data = np.random.rand(time, lat, lon)
    return xr.DataArray(data, coords=coords, dims=("time", "lat", "lon"))

def test_fit_and_predict():
    x = create_dummy_data()
    y = create_dummy_data()

    model = SamudraAI(time_seq=5)
    model.fit(x, y, epochs=1)  # cepat karena epoch 1

    corrected = model.predict(x)
    assert isinstance(corrected, xr.DataArray)
    assert "time" in corrected.dims



# src/samudra_ai/utils.py

import json
import numpy as np
import xarray as xr
import datetime
from sklearn.metrics import mean_absolute_error

def standardize_dims(ds: xr.DataArray) -> xr.DataArray:
    """Menyeragamkan nama dimensi ke 'lat' dan 'lon'."""
    var_name = ds.name or "variable"
    ds.name = var_name
    ds_temp = ds.to_dataset()
    rename_dict = {}
    for name in ['latitude', 'y', 'j']:
        if name in ds_temp.dims: rename_dict[name] = 'lat'
    for name in ['longitude', 'x', 'i']:
        if name in ds_temp.dims: rename_dict[name] = 'lon'
    da = ds_temp.rename(rename_dict)[var_name]
    # pastikan urutan time, lat, lon jika ada
    if all(dim in da.dims for dim in ["time", "lat", "lon"]):
        da = da.transpose("time", "lat", "lon")
    return da

def compute_metrics(obs, model):
    """Menghitung metrik statistik antara data observasi dan model."""
    if isinstance(obs, xr.DataArray): obs = obs.values
    if isinstance(model, xr.DataArray): model = model.values
    obs_vals, model_vals = obs.flatten(), model.flatten()
    mask = ~np.isnan(obs_vals) & ~np.isnan(model_vals)
    obs_clean, model_clean = obs_vals[mask], model_vals[mask]
    if len(obs_clean) == 0: return np.nan, np.nan, np.nan, np.nan
    correlation = np.corrcoef(obs_clean, model_clean)[0, 1]
    rmse = np.sqrt(np.mean((model_clean - obs_clean) ** 2))
    bias = np.mean(model_clean - obs_clean)
    mae = mean_absolute_error(obs_clean, model_clean)
    return correlation, rmse, bias, mae

def save_to_netcdf(da: xr.DataArray, path: str, attrs: dict = None):
    """
    Simpan xarray.DataArray ke file NetCDF dengan atribut dan encoding standar.
    """
    da.name = da.name or "variable"
    da.attrs.update(attrs or {})
    da.attrs.setdefault("title", "Hasil Koreksi SAMUDRA-AI")
    da.attrs.setdefault("description", "Data hasil koreksi bias menggunakan CNN-BiLSTM")
    da.attrs.setdefault("history", f"Generated by SAMUDRA-AI on {datetime.datetime.now().isoformat()}")
    
    # Tambahkan encoding agar tidak terlalu besar (opsional)
    encoding = {
        da.name: {
            "zlib": True,
            "complevel": 4,
            "_FillValue": -9999.0,
        }
    }
    da.to_netcdf(path, encoding=encoding, engine="h5netcdf")
    print(f"💾 Data berhasil disimpan ke: {path}")

class NumpyEncoder(json.JSONEncoder):
    """ Kelas helper untuk encoding NumPy array ke JSON. """
    def default(self, obj):
        if isinstance(obj, np.generic):
            return obj.item()
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super().default(obj)


# # File: src/samudra_ai/data_loader.py
# import os
# import xarray as xr
# import pandas as pd
# import numpy as np
# import cftime
# from .utils import standardize_dims

# def load_and_mask_dataset(file_path: str, var_name: str, lat_range: tuple, lon_range: tuple, time_range: tuple) -> xr.DataArray:
#     print(f"-> Memuat dan memproses file: {os.path.basename(file_path)}...")

#     # === Validasi time_range ===
#     if not (isinstance(time_range, (list, tuple)) and len(time_range) == 2):
#         raise ValueError("time_range harus tuple (start, end) dengan format YYYY-MM-DD")

#     try:
#         start_dt = pd.to_datetime(time_range[0])
#         end_dt = pd.to_datetime(time_range[1])
#     except Exception:
#         raise ValueError("Format waktu harus bisa dibaca oleh pandas, misal '1993-01-01'")

#     if start_dt >= end_dt:
#         raise ValueError("time_range: tanggal awal harus lebih kecil dari tanggal akhir")

#     # === Buka dataset dengan fallback engine ===
#     try:
#         data = xr.open_dataset(file_path, engine="h5netcdf", decode_times=True)
#     except Exception:
#         print(f"⚠️  Engine 'h5netcdf' gagal membaca {os.path.basename(file_path)}, mencoba 'netcdf4'...")
#         data = xr.open_dataset(file_path, engine="netcdf4", decode_times=True)

#     # === Validasi variabel ===
#     if var_name not in data.variables:
#         raise ValueError(f"Variabel '{var_name}' tidak ditemukan dalam {file_path}")

#     # === Deteksi nama dimensi waktu ===
#     time_candidates = ["time", "valid_time", "date", "Time", "t"]
#     detected_time = next((t for t in time_candidates if t in data.dims or t in data.coords), None)
#     if not detected_time:
#         raise ValueError("Dimensi waktu tidak ditemukan. Harus ada salah satu dari: time/valid_time/date/Time/t")

#     # Rename jadi "time" supaya konsisten
#     if detected_time != "time":
#         data = data.rename({detected_time: "time"})

#     if "time" not in data.coords:
#         raise ValueError("Koordinat 'time' tidak ditemukan.")

#     # === Penanganan cftime vs datetime64 ===
#     time_type = type(data.time.values[0])
#     if 'cftime' in str(time_type):
#         start_time, end_time = time_type(start_dt.year, start_dt.month, start_dt.day), time_type(end_dt.year, end_dt.month, end_dt.day)
#     else:
#         start_time, end_time = np.datetime64(start_dt, 'D'), np.datetime64(end_dt, 'D')

#     # === Slicing waktu ===
#     start_time_sel = data.time.sel(time=start_time, method="nearest").values
#     end_time_sel = data.time.sel(time=end_time, method="nearest").values
#     sliced_data = data[var_name].sel(time=slice(start_time_sel, end_time_sel))

#     # === Deteksi koordinat spasial ===
#     lat_names = ["lat", "latitude", "j", "y", "J", "Y"]
#     lon_names = ["lon", "longitude", "i", "x", "I", "X"]
#     detected_lat = next((lat for lat in lat_names if lat in sliced_data.dims), None)
#     detected_lon = next((lon for lon in lon_names if lon in sliced_data.dims), None)

#     if not detected_lat or not detected_lon:
#         raise ValueError("Dimensi lat/lon tidak ditemukan.")

#     # === Urutkan agar slicing aman ===
#     if sliced_data[detected_lat][0] > sliced_data[detected_lat][-1]:
#         sliced_data = sliced_data.sortby(detected_lat)
#     if sliced_data[detected_lon][0] > sliced_data[detected_lon][-1]:
#         sliced_data = sliced_data.sortby(detected_lon)

#     # === Slicing spasial ===
#     masked_data = sliced_data.sel(
#         {detected_lat: slice(*lat_range), detected_lon: slice(*lon_range)}
#     ).dropna(dim="time", how="all")

#     if masked_data.size == 0:
#         raise ValueError("Data kosong setelah slicing.")

#     # ✨ Standarisasi dimensi agar selalu lat/lon konsisten
#     return standardize_dims(masked_data)

# File: src/samudra_ai/data_loader.py yang baruuuuuuu
import os
import xarray as xr
import pandas as pd
import numpy as np
import cftime
from .utils import standardize_dims

def load_and_mask_dataset(file_path: str, var_name: str, lat_range: tuple, lon_range: tuple, time_range: tuple) -> xr.DataArray:
    """
    Memuat dataset NetCDF dan otomatis menyesuaikan nama dimensi spasial & waktu
    agar seragam [time, lat, lon]. Juga bisa menangani berbagai engine (h5netcdf/netCDF4)
    dan tipe koordinat (1D atau 2D).

    Parameters
    ----------
    file_path : str
        Path ke file NetCDF
    var_name : str
        Nama variabel utama
    lat_range, lon_range : tuple(float, float)
        Rentang spasial (misal: (-10, 10), (100, 120))
    time_range : tuple(str, str)
        Rentang waktu ("YYYY-MM-DD", "YYYY-MM-DD")
    """
    print(f"-> Memuat dan memproses file: {os.path.basename(file_path)}...")

    # === Validasi waktu ===
    if not (isinstance(time_range, (list, tuple)) and len(time_range) == 2):
        raise ValueError("time_range harus tuple (start, end) dengan format YYYY-MM-DD")

    start_dt, end_dt = map(pd.to_datetime, time_range)
    if start_dt >= end_dt:
        raise ValueError("Tanggal awal harus lebih kecil dari tanggal akhir")

    # === Buka dataset dengan fallback engine ===
    try:
        data = xr.open_dataset(file_path, engine="h5netcdf", decode_times=True)
    except Exception:
        print(f"⚠️ Engine 'h5netcdf' gagal, mencoba 'netCDF4'...")
        data = xr.open_dataset(file_path, engine="netcdf4", decode_times=True)

    # === Pastikan variabel ada ===
    if var_name not in data.variables:
        raise ValueError(f"Variabel '{var_name}' tidak ditemukan di {file_path}")

    da = data[var_name]

    # === Deteksi nama waktu ===
    time_candidates = ["time", "valid_time", "date", "Time", "t"]
    detected_time = next((t for t in time_candidates if t in data.dims or t in data.coords), None)
    if not detected_time:
        raise ValueError("Tidak menemukan koordinat waktu (time/valid_time/date/Time/t).")
    if detected_time != "time":
        data = data.rename({detected_time: "time"})
        da = data[var_name]

    # === Konversi waktu (cftime ke datetime64) ===
    if np.issubdtype(da.time.dtype, np.datetime64):
        pass
    else:
        try:
            da["time"] = xr.decode_cf(da).time
        except Exception:
            try:
                da["time"] = pd.to_datetime(da["time"].values)
            except Exception:
                print("⚠️ Waktu tidak bisa dikonversi otomatis, akan gunakan tipe original.")

    # === Seleksi waktu ===
    try:
        da = da.sel(time=slice(start_dt, end_dt))
    except Exception:
        # fallback jika cftime
        time_type = type(da.time.values[0])
        if "cftime" in str(time_type):
            da = da.sel(time=slice(
                time_type(start_dt.year, start_dt.month, start_dt.day),
                time_type(end_dt.year, end_dt.month, end_dt.day)
            ))

    if da.time.size == 0:
        raise ValueError("Tidak ada data dalam rentang waktu yang diberikan.")

    # === Deteksi nama lat/lon ===
    lat_names = ["lat", "latitude", "nav_lat", "y", "j", "LAT", "Y"]
    lon_names = ["lon", "longitude", "nav_lon", "x", "i", "LON", "X"]

    detected_lat = next((k for k in lat_names if k in da.dims or k in da.coords), None)
    detected_lon = next((k for k in lon_names if k in da.dims or k in da.coords), None)

    if not detected_lat or not detected_lon:
        raise ValueError("Koordinat lat/lon tidak ditemukan dalam dataset.")

    # === Tangani jika lat/lon berbentuk 2D (curvilinear) ===
    lat_vals = da[detected_lat]
    lon_vals = da[detected_lon]
    if lat_vals.ndim == 2 or lon_vals.ndim == 2:
        print("⚠️ Deteksi grid curvilinear (2D), slicing langsung akan di-skip.")
        sliced_data = da  # tidak bisa dislice langsung
    else:
        # Urutkan dan slice
        if lat_vals[0] > lat_vals[-1]:
            da = da.sortby(detected_lat)
        if lon_vals[0] > lon_vals[-1]:
            da = da.sortby(detected_lon)
        sliced_data = da.sel(
            {detected_lat: slice(*lat_range), detected_lon: slice(*lon_range)}
        )

    # === Hapus timestep kosong ===
    sliced_data = sliced_data.dropna(dim="time", how="all")

    if sliced_data.size == 0:
        raise ValueError("Data kosong setelah slicing spasial/waktu.")

    # === Rename dimensi agar konsisten ===
    rename_dict = {detected_lat: "lat", detected_lon: "lon"}
    if detected_lat != "lat" or detected_lon != "lon":
        sliced_data = sliced_data.rename(rename_dict)

    # === Standarisasi dimensi ===
    sliced_data = standardize_dims(sliced_data)

    print(f"✅ Dataset siap → shape={sliced_data.shape}, var={var_name}")
    return sliced_data



from typing import Tuple, List, Optional
import tensorflow as tf
from tensorflow.keras.layers import (
    Input, ConvLSTM2D, BatchNormalization, Conv2D, Lambda,
    SpatialDropout2D, SpatialDropout3D, TimeDistributed
)
from tensorflow.keras.models import Model
from tensorflow.keras.optimizers import Adam

class ModelConfig2:
    """Konfigurasi fleksibel untuk ConvLSTM2D Samudra-AI"""
    def __init__(
        self,
        conv_filters: List[int] = [16],         # filter untuk layer Conv2D awal
        lstm_filters: List[int] = [16, 32, 64, 128],  # filter tiap layer ConvLSTM2D
        kernel_size: Tuple[int, int] = (3, 3),
        activation: str = "relu",
        dropout_rates: List[float] = [0.1, 0.2, 0.3, 0.4],  # dropout sesuai layer ConvLSTM
        learning_rate: float = 1e-3,
        batch_size=8,
        epochs=100,      
    ):
        self.conv_filters = conv_filters
        self.lstm_filters = lstm_filters
        self.kernel_size = kernel_size
        self.activation = activation
        self.dropout_rates = dropout_rates
        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.epochs = epochs


def build_convlstm2d(
    input_shape: Tuple,
    output_shape: Tuple,
    config: Optional[ModelConfig2] = None
) -> Model:
    """Bangun arsitektur ConvLSTM2D sesuai baseline, tapi fleksibel."""
    if config is None:
        config = ModelConfig2()

    target_height, target_width = output_shape[0], output_shape[1]
    inputs = Input(shape=input_shape, name="input_layer")

    x = inputs

    # === Step 1: Conv2D awal ===
    for f in config.conv_filters:
        x = TimeDistributed(Conv2D(f, config.kernel_size, padding="same", activation=config.activation))(x)
        x = TimeDistributed(BatchNormalization())(x)

    # === Step 2: ConvLSTM stack ===
    for i, f in enumerate(config.lstm_filters):
        return_seq = i < len(config.lstm_filters) - 1
        x = ConvLSTM2D(f, config.kernel_size, padding="same",
                       return_sequences=return_seq,
                       activation=config.activation)(x)
        x = BatchNormalization()(x)

        if i < len(config.dropout_rates):
            if return_seq:
                x = TimeDistributed(SpatialDropout2D(config.dropout_rates[i]))(x)
            else:
                x = SpatialDropout2D(config.dropout_rates[i])(x)

    # === Resize ke target resolusi ===
    x = Lambda(lambda t: tf.image.resize(t, (target_height, target_width), method="bilinear"))(x)

    # === Conv2D final → output ===
    outputs = Conv2D(1, (1, 1), padding="same", activation="linear")(x)

    # === Compile ===
    model = Model(inputs, outputs)
    model.compile(optimizer=Adam(config.learning_rate), loss="mse", metrics=["mae"])

    return model


# # File: src/samudra_ai/models.py
from typing import Tuple, List, Optional
import numpy as np
from tensorflow.keras.models import Model
from tensorflow.keras.layers import (
    Input, TimeDistributed, Conv2D, BatchNormalization,
    Dropout, Flatten, Bidirectional, LSTM, Dense, Reshape, LeakyReLU
)
from tensorflow.keras.optimizers import Adam


class ModelConfig:
    """Konfigurasi fleksibel untuk CNN-BiLSTM Samudra-AI"""
    def __init__(
        self,
        filters: List[int] = [16, 32, 64, 128],
        kernel_size: Tuple[int, int] = (3, 3),
        activation: str = "linear",
        rnn_units: int = 64,
        dropout_rates: List[float] = [0.0, 0.1, 0.2, 0.0],
        leaky_relu_alpha: float = 0.1,
        batch_size: int = 32,
        epochs: int = 50,
        learning_rate: float = 1e-4
    ):
        self.filters = filters
        self.kernel_size = kernel_size
        self.activation = activation
        self.rnn_units = rnn_units
        self.dropout_rates = dropout_rates
        self.leaky_relu_alpha = leaky_relu_alpha
        self.batch_size = batch_size
        self.epochs = epochs
        self.learning_rate = learning_rate


def build_cnn_bilstm(
    input_shape: Tuple,
    output_shape: Tuple,
    config: Optional[ModelConfig] = None
) -> Model:
    """
    Bangun arsitektur CNN-BiLSTM sesuai baseline,
    tapi fleksibel untuk variabel harian/bulanan dan berbagai parameter.
    """
    if config is None:
        config = ModelConfig()  # pakai default baseline

    target_height, target_width = output_shape[0], output_shape[1]
    inputs = Input(shape=input_shape, name="input_layer")
    x = inputs

    # === TimeDistributed CNN layers ===
    for i, f in enumerate(config.filters):
        x = TimeDistributed(
            Conv2D(f, config.kernel_size, padding="same", activation=config.activation)
        )(x)
        x = TimeDistributed(LeakyReLU(config.leaky_relu_alpha))(x)
        x = TimeDistributed(BatchNormalization())(x)

        # Tambahkan dropout sesuai urutan kalau tersedia
        if i < len(config.dropout_rates) and config.dropout_rates[i] > 0:
            x = TimeDistributed(Dropout(config.dropout_rates[i]))(x)

    # === Flatten dan Dropout ===
    x = TimeDistributed(Flatten())(x)
    x = Dropout(config.dropout_rates[-1] if len(config.dropout_rates) > 0 else 0.2)(x)

    # === BiLSTM layers ===
    x = Bidirectional(LSTM(config.rnn_units, return_sequences=True))(x)
    x = Bidirectional(LSTM(config.rnn_units, return_sequences=False))(x)

    # === Dense ke output ===
    x = Dense(target_height * target_width)(x)
    outputs = Reshape((target_height, target_width, 1))(x)

    # === Compile model ===
    model = Model(inputs=inputs, outputs=outputs)
    model.compile(
        optimizer=Adam(learning_rate=config.learning_rate),
        loss="mse",
        metrics=["mae"]
    )

    return model
